@@ -41,27 +41,13 @@
 #include "gnome-settings-profile.h"
 #include "gsd-keybindings-manager.h"
 
+#include "gsd-keygrab.h"
 #include "eggaccelerators.h"
 
-/* we exclude shift, GDK_CONTROL_MASK and GDK_MOD1_MASK since we know what
-   these modifiers mean
-   these are the mods whose combinations are bound by the keygrabbing code */
-#define IGNORED_MODS (0x2000 /*Xkb modifier*/ | GDK_LOCK_MASK  | \
-        GDK_MOD2_MASK | GDK_MOD3_MASK | GDK_MOD4_MASK | GDK_MOD5_MASK)
-/* these are the ones we actually use for global keys, we always only check
- * for these set */
-#define USED_MODS (GDK_SHIFT_MASK | GDK_CONTROL_MASK | GDK_MOD1_MASK)
-
 #define GCONF_BINDING_DIR "/desktop/gnome/keybindings"
 
 #define GSD_KEYBINDINGS_MANAGER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GSD_TYPE_KEYBINDINGS_MANAGER, GsdKeybindingsManagerPrivate))
 
-typedef struct {
-        guint keysym;
-        guint state;
-        guint keycode;
-} Key;
-
 typedef struct {
         char *binding_str;
         char *action;
@@ -74,6 +60,7 @@ struct GsdKeybindingsManagerPrivate
 {
         GSList *binding_list;
         GSList *screens;
+        guint   notify;
 };
 
 static void     gsd_keybindings_manager_class_init  (GsdKeybindingsManagerClass *klass);
@@ -169,11 +156,9 @@ bindings_get_entry (GsdKeybindingsManager *manager,
         char    *gconf_key;
         char    *action = NULL;
         char    *key = NULL;
-        gboolean ret = FALSE;
 
         g_return_val_if_fail (subdir != NULL, FALSE);
 
-        /* value = gconf_entry_get_value (entry); */
         gconf_key = g_path_get_basename (subdir);
 
         if (!gconf_key) {
@@ -188,29 +173,25 @@ bindings_get_entry (GsdKeybindingsManager *manager,
                 char *key_name = g_path_get_basename (gconf_entry_get_key (entry));
 
                 if (key_name == NULL) {
-                        goto out;
+                        /* ignore entry */
                 } else if (strcmp (key_name, "action") == 0) {
-                        if (!action) {
-                                action = entry_get_string (entry);
-                        } else {
-                                g_warning (_("Key Binding (%s) has its action defined multiple times\n"),
-                                           gconf_key);
-                        }
+                        action = entry_get_string (entry);
                 } else if (strcmp (key_name, "binding") == 0) {
-                        if (!key) {
-                                key = entry_get_string (entry);
-                        } else {
-                                g_warning (_("Key Binding (%s) has its binding defined multiple times\n"),
-                                           gconf_key);
-                        }
+                        key = entry_get_string (entry);
                 }
 
+                g_free (key_name);
                 gconf_entry_free (entry);
         }
 
+        g_slist_free (list);
+
         if (!action || !key) {
-                g_warning (_("Key Binding (%s) is incomplete\n"), gconf_key);
-                goto out;
+                g_warning (_("Key binding (%s) is incomplete"), gconf_key);
+                g_free (gconf_key);
+                g_free (action);
+                g_free (key);
+                return FALSE;
         }
 
         tmp_elem = g_slist_find_custom (manager->priv->binding_list,
@@ -220,9 +201,10 @@ bindings_get_entry (GsdKeybindingsManager *manager,
         if (!tmp_elem) {
                 new_binding = g_new0 (Binding, 1);
         } else {
-                new_binding = (Binding*) tmp_elem->data;
+                new_binding = (Binding *) tmp_elem->data;
                 g_free (new_binding->binding_str);
                 g_free (new_binding->action);
+                g_free (new_binding->gconf_key);
         }
 
         new_binding->binding_str = key;
@@ -234,20 +216,23 @@ bindings_get_entry (GsdKeybindingsManager *manager,
         new_binding->previous_key.keycode = new_binding->key.keycode;
 
         if (parse_binding (new_binding)) {
-                manager->priv->binding_list = g_slist_append (manager->priv->binding_list, new_binding);
-                ret = TRUE;
+                if (!tmp_elem)
+                        manager->priv->binding_list = g_slist_prepend (manager->priv->binding_list, new_binding);
         } else {
-                g_warning (_("Key Binding (%s) is invalid\n"), gconf_key);
+                g_warning (_("Key binding (%s) is invalid"), gconf_key);
                 g_free (new_binding->binding_str);
                 g_free (new_binding->action);
+                g_free (new_binding->gconf_key);
+                g_free (new_binding);
+
+                if (tmp_elem)
+                        manager->priv->binding_list = g_slist_delete_link (manager->priv->binding_list, tmp_elem);
+                return FALSE;
         }
 
- out:
-        g_slist_free (list);
-        return ret;
+        return TRUE;
 }
 
-
 static gboolean
 key_already_used (GsdKeybindingsManager *manager,
                   Binding               *binding)
@@ -266,79 +251,6 @@ key_already_used (GsdKeybindingsManager *manager,
         return FALSE;
 }
 
-static void
-grab_key (GdkWindow *root,
-          Key       *key,
-          int        result,
-          gboolean   grab)
-{
-        gdk_error_trap_push ();
-
-        if (grab) {
-                XGrabKey (GDK_DISPLAY (),
-                          key->keycode,
-                          (result | key->state),
-                          GDK_WINDOW_XID (root),
-                          True,
-                          GrabModeAsync,
-                          GrabModeAsync);
-        } else {
-                XUngrabKey (GDK_DISPLAY (),
-                            key->keycode,
-                            (result | key->state),
-                            GDK_WINDOW_XID (root));
-        }
-        gdk_flush ();
-        if (gdk_error_trap_pop ()) {
-                g_warning (_("It seems that another application already has access to key '%u'."),
-                           key->keycode);
-        }
-}
-
-/* inspired from all_combinations from gnome-panel/gnome-panel/global-keys.c */
-#define N_BITS 32
-static void
-do_grab (GsdKeybindingsManager *manager,
-         gboolean               grab,
-         Key                   *key)
-{
-        int   indexes[N_BITS];/*indexes of bits we need to flip*/
-        int   i;
-        int   bit;
-        int   bits_set_cnt;
-        int   uppervalue;
-        guint mask_to_traverse = IGNORED_MODS & ~ key->state;
-
-        bit = 0;
-        for (i = 0; i < N_BITS; i++) {
-                if (mask_to_traverse & (1<<i)) {
-                        indexes[bit++]=i;
-                }
-        }
-
-        bits_set_cnt = bit;
-
-        uppervalue = 1 << bits_set_cnt;
-        for (i = 0; i < uppervalue; i++) {
-                GSList *l;
-                int j, result = 0;
-
-                for (j = 0; j < bits_set_cnt; j++) {
-                        if (i & (1<<j)) {
-                                result |= (1<<indexes[j]);
-                        }
-                }
-
-                for (l = manager->priv->screens; l ; l = l->next) {
-                  GdkScreen *screen = l->data;
-                  grab_key (gdk_screen_get_root_window (screen),
-                            key,
-                            result,
-                            grab);
-                }
-        }
-}
-
 static void
 binding_register_keys (GsdKeybindingsManager *manager)
 {
@@ -355,15 +267,15 @@ binding_register_keys (GsdKeybindingsManager *manager)
                         /* Ungrab key if it changed and not clashing with previously set binding */
                         if (! key_already_used (manager, binding)) {
                                 if (binding->previous_key.keycode) {
-                                        do_grab (manager, FALSE, &binding->previous_key);
+                                        grab_key (&binding->previous_key, FALSE, manager->priv->screens);
                                 }
-                                do_grab (manager, TRUE, &binding->key);
+                                grab_key (&binding->key, TRUE, manager->priv->screens);
 
                                 binding->previous_key.keysym = binding->key.keysym;
                                 binding->previous_key.state = binding->key.state;
                                 binding->previous_key.keycode = binding->key.keycode;
                         } else
-                                g_warning (_("Key Binding (%s) is already in use\n"), binding->binding_str);
+                                g_warning ("Key binding (%s) is already in use", binding->binding_str);
                 }
         }
         gdk_flush ();
@@ -454,23 +366,17 @@ keybindings_filter (GdkXEvent             *gdk_xevent,
                     GdkEvent              *event,
                     GsdKeybindingsManager *manager)
 {
-        XEvent *xevent = (XEvent *)gdk_xevent;
-        guint   keycode;
-        guint   state;
+        XEvent *xevent = (XEvent *) gdk_xevent;
         GSList *li;
 
         if (xevent->type != KeyPress) {
                 return GDK_FILTER_CONTINUE;
         }
 
-        keycode = xevent->xkey.keycode;
-        state = xevent->xkey.state;
-
         for (li = manager->priv->binding_list; li != NULL; li = li->next) {
-                Binding *binding = (Binding*) li->data;
+                Binding *binding = (Binding *) li->data;
 
-                if (keycode == binding->key.keycode &&
-                    (state & USED_MODS) == binding->key.state) {
+                if (match_key (&binding->key, xevent)) {
                         GError  *error = NULL;
                         gboolean retval;
                         gchar  **argv = NULL;
@@ -541,19 +447,14 @@ bindings_callback (GConfClient           *client,
         binding_register_keys (manager);
 }
 
-static void
+static guint
 register_config_callback (GsdKeybindingsManager   *manager,
+                          GConfClient             *client,
                           const char              *path,
                           GConfClientNotifyFunc    func)
 {
-        GConfClient *client;
-
-        client = gconf_client_get_default ();
-
         gconf_client_add_dir (client, path, GCONF_CLIENT_PRELOAD_NONE, NULL);
-        gconf_client_notify_add (client, path, func, manager, NULL, NULL);
-
-        g_object_unref (client);
+        return gconf_client_notify_add (client, path, func, manager, NULL, NULL);
 }
 
 gboolean
@@ -574,19 +475,20 @@ gsd_keybindings_manager_start (GsdKeybindingsManager *manager,
         dpy = gdk_display_get_default ();
         screen_num = gdk_display_get_n_screens (dpy);
 
-        register_config_callback (manager,
-                                  GCONF_BINDING_DIR,
-                                  (GConfClientNotifyFunc)bindings_callback);
-
         for (i = 0; i < screen_num; i++) {
                 screen = gdk_display_get_screen (dpy, i);
                 gdk_window_add_filter (gdk_screen_get_root_window (screen),
-                                       (GdkFilterFunc)keybindings_filter,
+                                       (GdkFilterFunc) keybindings_filter,
                                        manager);
         }
 
         client = gconf_client_get_default ();
 
+        manager->priv->notify = register_config_callback (manager,
+                                                          client,
+                                                          GCONF_BINDING_DIR,
+                                                          (GConfClientNotifyFunc) bindings_callback);
+
         list = gconf_client_all_dirs (client, GCONF_BINDING_DIR, NULL);
         manager->priv->screens = get_screens_list ();
 
@@ -608,7 +510,37 @@ gsd_keybindings_manager_start (GsdKeybindingsManager *manager,
 void
 gsd_keybindings_manager_stop (GsdKeybindingsManager *manager)
 {
+        GsdKeybindingsManagerPrivate *p = manager->priv;
+        GSList *l;
+
         g_debug ("Stopping keybindings manager");
+
+        if (p->notify != 0) {
+                GConfClient *client = gconf_client_get_default ();
+                gconf_client_remove_dir (client, GCONF_BINDING_DIR, NULL);
+                gconf_client_notify_remove (client, p->notify);
+                g_object_unref (client);
+                p->notify = 0;
+        }
+
+        for (l = p->screens; l; l = l->next) {
+                GdkScreen *screen = l->data;
+                gdk_window_remove_filter (gdk_screen_get_root_window (screen),
+                                          (GdkFilterFunc) keybindings_filter,
+                                          manager);
+        }
+        g_slist_free (p->screens);
+        p->screens = NULL;
+
+        for (l = p->binding_list; l; l = l->next) {
+                Binding *b = l->data;
+                g_free (b->binding_str);
+                g_free (b->action);
+                g_free (b->gconf_key);
+                g_free (b);
+        }
+        g_slist_free (p->binding_list);
+        p->binding_list = NULL;
 }
 
 static void
